@@ -5,7 +5,7 @@
  *
  *-----------------------------------------------------------------------------
  *
- *	Copyright (C) 2001-2006 Eric Thibaut.
+ *	Copyright (C) 2001-2007 Eric Thibaut.
  *
  *	This file is part of Yeti.
  *
@@ -26,8 +26,17 @@
  *-----------------------------------------------------------------------------
  *
  * History:
- *	$Id: yeti_hash.c,v 1.6 2006/07/19 14:44:35 eric Exp $
+ *	$Id: yeti_hash.c,v 1.7 2007/03/23 11:53:53 eric Exp eric $
  *	$Log: yeti_hash.c,v $
+ *	Revision 1.7  2007/03/23 11:53:53  eric
+ *	 - Hash table object can now have their own evaluator, which can be
+ *	   queried/set by the `h_evaluator` function.
+ *	 - New function `h_number` to query number of entries in a hash table.
+ *	 - Function `is_hash` returns 2 for a hash table object implementing
+ *	   its own evaluator.
+ *	 - h_clone, h_copy, h_info and h_show fixed to account for the
+ *	   evaluator of the object.
+ *
  *	Revision 1.6  2006/07/19 14:44:35  eric
  *	Copyright notice updated.
  *
@@ -62,6 +71,8 @@
 #include "yeti.h"
 #include "yio.h"
 
+#undef H_DEBUG
+
 /*---------------------------------------------------------------------------*/
 /* DEFINITIONS FOR STRING HASH TABLES */
 
@@ -93,6 +104,7 @@ typedef struct h_entry h_entry_t;
 struct h_table {
   int references;         /* reference counter */
   Operations *ops;        /* virtual function table */
+  long        eval;       /* index to eval method (-1L if none) */
   h_uint_t    number;     /* number of entries */
   h_uint_t    size;       /* number of allocated slots */
   h_uint_t    mask;       /* size-1 */
@@ -108,6 +120,10 @@ struct h_entry {
 			    whole string name to fit (MUST BE LAST MEMBER) */
 };
 
+static h_uint_t h_code[256]; /* array of integer code to check consistency
+				of a symbol's name or, if YETI_HASH_RESTRICT
+				is defined, to compute a hash key */
+
 #ifdef YETI_HASH_RESTRICT
 # define H_CODE(BYTE)    (h_code[BYTE])
 #else
@@ -175,6 +191,9 @@ extern int h_insert(h_table_t *table, const char *name, Symbol *sym);
 	(hence a new entry was created); 1 if a former entry in TABLE matched
 	NAME (which was properly unreferenced); -1 in case of error. */
 
+static void h_init(void);
+/*----- Initialize internals of hash table manager. */
+
 /*---------------------------------------------------------------------------*/
 /* PRIVATE ROUTINES */
 
@@ -197,6 +216,12 @@ static void get_member(Symbol *owner, h_table_t *table, const char *name);
 /*----- Replace stack symbol OWNER by the contents of entry matching NAME
 	in hash TABLE (taking care of UnRef/Ref properly). */
 
+static long get_method_index(Symbol *s, long defval);
+/*----- Return index in globTab of contents of symbol S which must be
+	a symbols'name or a function.  If S is the void symbol or a
+	NULL string, DEFVAL is returned. If the symbol is invalid, -1L
+	is returned. */
+
 #ifdef YETI_AVOID_CYCLIC_REFERENCES
 static void assert_no_cyclic_references(DataBlock *self, DataBlock *obj);
 /*----- Call YError if OBJ contains a reference to SELF.   All hash/table
@@ -218,6 +243,8 @@ static UnaryOp PrintH;
 static void FreeH(void *addr);  /* ******* Use Unref(hash) ******* */
 static void EvalH(Operand *op);
 
+static long default_eval_index = -1; /* index of default eval method in globTab */
+
 Operations hashOps = {
   &FreeH, T_OPAQUE, 0, /* promoteID = */T_STRING/* means illegal */,
   "hash_table",
@@ -242,7 +269,15 @@ static void PrintH(Operand *op)
   ForceNewline();
   PrintFunc("Object of type: ");
   PrintFunc(obj->ops->typeName);
-  sprintf(line, " (references=%d, number=%u, size=%u, mask=0x%x)",
+  PrintFunc(" (evaluator=");
+  if (obj->eval < 0L) {
+    PrintFunc("(nil)");
+  } else {
+    PrintFunc("\"");
+    PrintFunc(globalTable.names[obj->eval]);
+    PrintFunc("\"");
+  }
+  sprintf(line, ", references=%d, number=%u, size=%u, mask=0x%x)",
 	  obj->references, obj->number, obj->size, obj->mask);
   PrintFunc(line);
   ForceNewline();
@@ -259,22 +294,82 @@ static void EvalH(Operand *op)
 {
   char *name;
   long index = -1L;
-  Symbol *stack = op->owner;
-  Symbol *sym;
+  Symbol *s, *owner;
   h_table_t *table;
   h_entry_t *entry = NULL;
-  DataBlock *old;
+  DataBlock *old, *db;
   OpTable *ops;
+  Operations *oper;
+  int i, nargs, offset;
  
-  if (sp - stack == 1) {
-    /* got exactly one argument */
+  /* first get the hash table */
+  owner = op->owner;
+  nargs = sp - owner; /* number of arguments */
+#if defined(H_DEBUG) && (H_DEBUG >= 1)
+  s = owner;
+  if (s->ops != &dataBlockSym) {
+    fprintf(stderr, "unexpected non-DB in Eval method\n");
+    while (s->ops == &referenceSym) {
+      s = &globTab[owner->index];
+    }
+  }
+  if (s->ops != &dataBlockSym || s->value.db->ops != &hashOps) {
+    YError("unexpected non-hash table object (must be a BUG!)");
+  }
+  table = (h_table_t *)s->value.db;
+#else
+  table = (h_table_t *)owner->value.db;
+#endif
 
-    /* first get the hash table */
-    sym = (stack->ops == &referenceSym) ? &globTab[stack->index] : stack;
-    if (sym->ops != &dataBlockSym || sym->value.db->ops != &hashOps)
-      YError("unexpected non-hash table object (must be a BUG!)");
-    table = (h_table_t *)sym->value.db;
-    
+  if (table->eval >= 0L) {
+    /* this hash table implement its own eval method */
+    s = &globTab[table->eval];
+    while (s->ops == &referenceSym) {
+      s = &globTab[s->index];
+    }
+    db = s->value.db; /* correctness checked below */
+    if (s->ops != &dataBlockSym || ((oper = db->ops) != &functionOps &&
+				    oper != &builtinOps &&
+				    oper != &auto_ops)) {
+      YError("non-function eval method");
+    }
+
+    /* shift stack to prepend reference to eval method */
+    offset = owner - spBottom; /* stack may move */
+    if (CheckStack(2)) {
+      owner = spBottom + offset;
+      op->owner = owner;
+    }
+    /*** CRITICAL CODE START ***/
+    {
+      volatile Symbol *stack = owner;
+      ++nargs; /* one more argument: the object itself */
+      i = nargs;
+      stack[i].ops = &intScalar; /* set safe OpTable */
+      sp = (Symbol *)stack + i; /* it is now safe to grow the stack */
+      while (--i >= 0) {
+	ops = stack[i].ops;
+	stack[i].ops = &intScalar; /* set safe OpTable */
+	stack[i + 1].value = stack[i].value;
+	stack[i + 1].index = stack[i].index;
+	stack[i + 1].ops = ops; /* set true OpTable *after* proper initialization */
+      }
+      stack->value.db = Ref(db);
+      stack->ops = &dataBlockSym;
+    }
+    /*** CRITICAL CODE END ***/
+
+    /* re-form operand and call Eval method */
+    op->owner = owner; /* stack may have moved */
+    op->references = nargs;   /* (see FormEvalOp in array.c) */
+    op->ops = db->ops;
+    op->value = db;
+    op->ops->Eval(op);
+    return;
+  }
+
+  /* got exactly one argument */
+  if (nargs == 1) {
     /* parse the argument */
     if (sp->ops == &longScalar) {
       index = sp->value.l;
@@ -331,22 +426,22 @@ static void EvalH(Operand *op)
 
  replace:
   Drop(1); /* discard key name or index (after using it) */
-  old = (stack->ops == &dataBlockSym) ? stack->value.db : NULL;
-  stack->ops = &intScalar;      /* avoid clash in case of interrupts */
+  old = (owner->ops == &dataBlockSym) ? owner->value.db : NULL;
+  owner->ops = &intScalar;      /* avoid clash in case of interrupts */
   if (entry) {
     if ((ops = entry->sym_ops) == &dataBlockSym) {
       DataBlock *db = entry->sym_value.db;
-      stack->value.db = Ref(db);
+      owner->value.db = Ref(db);
     } else {
-      stack->value = entry->sym_value;
+      owner->value = entry->sym_value;
     }
   } else {
     /* NULLER_DATA_BLOCK NewRange(0L, 0L, 1L, R_NULLER); */
-    stack->value.db = RefNC(&nilDB);
+    owner->value.db = RefNC(&nilDB);
     ops = &dataBlockSym;
   }
   Unref(old);
-  stack->ops = ops;           /* change ops only AFTER value updated */
+  owner->ops = ops;           /* change ops only AFTER value updated */
 }
 
 /*---------------------------------------------------------------------------*/
@@ -373,7 +468,15 @@ void Y_is_hash(int nargs)
   int result;
   if (nargs != 1) YError("is_hash takes exactly one argument");
   s = YETI_DEREF_SYMBOL(sp);
-  result = (s->ops == &dataBlockSym && s->value.db->ops == &hashOps);
+  if (s->ops == &dataBlockSym && s->value.db->ops == &hashOps) {
+    if (((h_table_t *)s->value.db)->eval >= 0L) {
+      result = 2;
+    } else {
+      result = 1;
+    }
+  } else {
+    result = 0;
+  }
   PushIntValue(result);
 }
 
@@ -494,6 +597,20 @@ void Y_h_pop(int nargs)
   PushDataBlock(RefNC(&nilDB));
 }
 
+void Y_h_number(int nargs)
+{
+  Symbol *s;
+  long result;
+
+  if (nargs != 1) YError("h_number takes exactly one argument");
+  s = YETI_DEREF_SYMBOL(sp);
+  if (s->ops != &dataBlockSym || s->value.db->ops != &hashOps) {
+    YError("inexpected non-hash table argument");
+  }
+  result = ((h_table_t *)s->value.db)->number;
+  PushLongValue(result);
+}
+
 void Y_h_keys(int nargs)
 {
   h_entry_t *entry;
@@ -625,6 +742,98 @@ void Y_h_stat(int nargs)
 #endif
 }
 
+void Y_h_evaluator(int nargs)
+{
+  h_table_t *table;
+  char *str;
+  long new_index, old_index;
+  int push_result;
+
+  if (nargs < 1 || nargs > 2) YError("h_evaluator takes 1 or 2 arguments");
+  push_result =  ! yarg_subroutine();
+  table = get_hash(sp - nargs + 1);
+  old_index = table->eval;
+
+  if (nargs == 2) {
+    new_index = get_method_index(sp, default_eval_index);
+    if (new_index < 0L) {
+      YError("evaluator must be a function or a valid symbol's name");
+    }
+    if (new_index == default_eval_index) {
+      table->eval = -1L;
+    } else {
+      table->eval = new_index;
+    }
+  }
+  if (push_result) {
+    if (old_index >= 0L && old_index != default_eval_index) {
+      str = globalTable.names[old_index];
+    } else {
+      str = (char *)0;
+    }
+    push_string_value(str);
+  }
+}
+
+#if YETI_MUST_DEFINE_AUTOLOAD_TYPE
+typedef struct autoload_t autoload_t;
+struct autoload_t {
+  int references;      /* reference counter */
+  Operations *ops;     /* virtual function table */
+  long ifile;          /* index into table of autoload files */
+  long isymbol;        /* global symtab index */
+  autoload_t *next;    /* linked list for each ifile */
+};
+#endif /* YETI_MUST_DEFINE_AUTOLOAD_TYPE */
+
+static long get_method_index(Symbol *s, long defval)
+{
+  Operations *ops;
+  unsigned char *str;
+  int i;
+  h_uint_t c;
+
+  while (s->ops == &referenceSym) {
+    s = &globTab[s->index];
+  }
+  if (s->ops == &dataBlockSym) {
+    ops = s->value.db->ops;
+    if (ops == &functionOps) {
+      return ((Function *)s->value.db)->code[0].index;
+    } else if (ops == &builtinOps) {
+      return ((BIFunction *)s->value.db)->index;
+    } else if (ops == &auto_ops) {
+      return ((autoload_t *)s->value.db)->isymbol;
+    } else if (ops == &stringOps) {
+      if (((Array *)s->value.db)->type.dims) {
+	return -1L;
+      }
+      str = (unsigned char *)((Array *)s->value.db)->value.q[0];
+      if (! str) {
+	/* nil symbol's name corresponds to default value */
+	return defval;
+      }
+      i = 0;
+      if (h_code[str[i]] <= 10U) {
+	/* symbol's name must not have a zero length, nor start with
+	   an invalid character nor a digit */
+	return -1L;
+      }
+      while ((c = str[++i])) {
+	if (! h_code[c]) {
+	  /* symbol's must not contain an invalid character */
+	  return -1L;
+	}
+      }
+      return Globalize((char *)str, i);
+    } else if (ops == &voidOps) {
+      /* void symbol corresponds to default value */
+      return defval;
+    }
+  }
+  return -1L;
+}
+
 /*---------------------------------------------------------------------------*/
 
 static void get_member(Symbol *owner, h_table_t *table, const char *name)
@@ -765,14 +974,6 @@ static void assert_no_cyclic_references(DataBlock *self, DataBlock *obj)
    algorithm is taken from Tcl (which is 25-30% more efficient than
    Yorick's algorithm). */
 
-#ifdef YETI_HASH_RESTRICT
-static h_uint_t h_code[256];
-static int first_time = 1;
-
-static void h_init(void);
-/*----- Initialize internals of hash table manager.  Does nothing, if compiled
-	with no restriction on hash key names. */
-
 static void h_init(void)
 {
   h_uint_t i, code = 0;
@@ -782,20 +983,17 @@ static void h_init(void)
   h_code['_'] = ++code;
   for (i='a' ; i<='z' ; ++i) h_code[i] = ++code;
 }
-#endif
 
 h_table_t *h_new(h_uint_t number)
 {
   h_uint_t nbytes, size = 1;
   h_table_t *table;
 
-#ifdef YETI_HASH_RESTRICT
   /* Initialization of internals. */
-  if (first_time) {
+  if (default_eval_index < 0L) {
     h_init();
-    first_time = 0;
+    default_eval_index = Globalize("*hash_evaluator*", 0L);
   }
-#endif
 
   /* Member SIZE of a hash table is always a power of 2, greater or
      equal 2*NUMBER (twice the number of entries in the table). */
@@ -814,6 +1012,7 @@ h_table_t *h_new(h_uint_t number)
   memset(table->slot, 0, nbytes);
   table->references = 0;
   table->ops = &hashOps;
+  table->eval = -1L;
   table->number = 0;
   table->size = size;
   table->mask = size - 1;
@@ -1031,4 +1230,11 @@ int h_insert(h_table_t *table, const char *sname, Symbol *sym)
   return 0; /* a new entry was created */
 }
 
-/*---------------------------------------------------------------------------*/
+/*---------------------------------------------------------------------------*
+ * Local Variables:                                                          *
+ * mode: C                                                                   *
+ * tab-width: 8                                                              *
+ * fill-column: 75                                                           *
+ * coding: latin-1                                                           *
+ * End:                                                                      *
+ *---------------------------------------------------------------------------*/
