@@ -26,8 +26,18 @@
  *-----------------------------------------------------------------------------
  *
  * History:
- *	$Id: yeti_misc.c,v 1.3 2006/07/19 17:33:11 eric Exp $
+ *	$Id: yeti_misc.c,v 1.6 2007/05/01 20:23:39 eric Exp $
  *	$Log: yeti_misc.c,v $
+ *	Revision 1.6  2007/05/01 20:23:39  eric
+ *	 - Fixed some compiler warnings.
+ *
+ *	Revision 1.5  2007/04/24 07:58:07  eric
+ *	 - The `symbol_names` function can now specifically select
+ *	   lists, hash tables and/or auto-loaded functions.
+ *
+ *	Revision 1.4  2006/12/05 07:19:51  eric
+ *	 - Renamed built-in `typeIDof` as `identof`.
+ *
  *	Revision 1.3  2006/07/19 17:33:11  eric
  *	 - New built-in function insure_temporary.
  *
@@ -45,6 +55,7 @@
 #include <ctype.h>
 #include <float.h>
 
+#include <yapi.h>
 #include <yio.h>
 #include <pstdio.h>
 #include <hlevel.h>
@@ -74,7 +85,7 @@ extern BuiltIn Y_symbol_exists, Y_symbol_names;
 extern BuiltIn Y_unref, Y_swap;
 extern BuiltIn Y_get_includes, Y_current_include;
 extern BuiltIn Y_filepath;
-extern BuiltIn Y_typeIDof, Y_nrefsof;
+extern BuiltIn Y_identof, Y_nrefsof;
 extern BuiltIn Y_smooth3;
 #if PROVIDE_IS_LIST
 extern BuiltIn Y_is_list;
@@ -998,11 +1009,11 @@ void Y_machine_constant(int argc)
 /*---------------------------------------------------------------------------*/
 /* SYMBOLS */
 
-void Y_typeIDof(int argc)
+void Y_identof(int argc)
 {
   Symbol *s;
   long typeID;
-  if (argc != 1) YError("typeIDof takes exactly one argument");
+  if (argc != 1) YError("identof takes exactly one argument");
   s = YETI_DEREF_SYMBOL(sp);
   if (s->ops==&doubleScalar) {
     typeID = T_DOUBLE;
@@ -1041,71 +1052,116 @@ void Y_symbol_exists(int argc)
 #define GET_STRUCTDEF  64
 #define GET_STREAM    128
 #define GET_OPAQUE    256
+#define GET_LIST      512
+#define GET_HASH     1024
+#define GET_AUTOLOAD 2048
+
 void Y_symbol_names(int argc)
 {
-  int flags;
+  extern Operations hashOps, listOps;
+  long i, nitems, number;
+  Operations *ops;
   char **ret;
-  long i, nitems;
   int match[T_OPAQUE+1];
+  int type, flags, pass, get_list, get_hash, get_autoload, get_other;
 
   if (argc != 1) YError("symbol_list takes exactly one argument");
   flags = yeti_get_optional_integer(sp, (GET_ARRAY | GET_STRUCT | GET_RANGE |
 					 GET_FUNCTION | GET_BUILTIN |
 					 GET_STRUCTDEF | GET_STREAM |
 					 GET_OPAQUE));
-  /*Drop(argc+1);*/
   nitems = globalTable.nItems;
-  if (nitems > 0) {
-    if (flags == -1) {
-      ret = YETI_PUSH_NEW_Q(ynew_dim(nitems, NULL));
-      for (i=0 ; i<nitems ; ++i) {
-	ret[i] = p_strcpy(globalTable.names[i]);
-      }
-      return;
-    } else {
-      unsigned int typeID;
-      long number;
-      for (i=0 ; i<=T_OPAQUE ; ++i) match[i] = 0;
-      match[T_CHAR]      = (flags&GET_ARRAY);
-      match[T_SHORT]     = (flags&GET_ARRAY);
-      match[T_INT]       = (flags&GET_ARRAY);
-      match[T_LONG]      = (flags&GET_ARRAY);
-      match[T_FLOAT]     = (flags&GET_ARRAY);
-      match[T_DOUBLE]    = (flags&GET_ARRAY);
-      match[T_COMPLEX]   = (flags&GET_ARRAY);
-      match[T_STRING]    = (flags&GET_ARRAY);
-      match[T_POINTER]   = (flags&GET_ARRAY);
-      match[T_STRUCT]    = (flags&GET_STRUCT);
-      match[T_RANGE]     = (flags&GET_RANGE);
-#ifdef GET_LVALUE
-      match[T_LVALUE]    = (flags&GET_LVALUE);
-#endif
-      match[T_VOID]      = (flags&GET_VOID);
-      match[T_FUNCTION]  = (flags&GET_FUNCTION);
-      match[T_BUILTIN]   = (flags&GET_BUILTIN);
-      match[T_STRUCTDEF] = (flags&GET_STRUCTDEF);
-      match[T_STREAM]    = (flags&GET_STREAM);
-      match[T_OPAQUE]    = (flags&GET_OPAQUE);
-
-#define MATCH(i) (globTab[i].ops == &dataBlockSym \
-                  && (typeID = globTab[i].value.db->ops->typeID) <= T_OPAQUE \
-	          && match[typeID])
-      number = 0;
-      for (i=0 ; i<nitems ; ++i) {
-	if (MATCH(i)) ++number;
-      }
-      if (number) {
-	ret = YETI_PUSH_NEW_Q(ynew_dim(number, NULL));
-	for (i=0 ; i<nitems ; ++i) {
-	  if (MATCH(i)) *ret++ = p_strcpy(globalTable.names[i]);
-	}
-	return;
-      }
-#undef MATCH
+  if (nitems <= 0) {
+    /* No symbols defined. */
+    PushDataBlock(RefNC(&nilDB));
+    return;
+  }
+  if (flags == -1) {
+    /* Return names of all symbols ever defined. */
+    ret = YETI_PUSH_NEW_Q(ynew_dim(nitems, NULL));
+    for (i=0 ; i<nitems ; ++i) {
+      ret[i] = p_strcpy(globalTable.names[i]);
+    }
+    return;
+  }
+  if (flags & GET_OPAQUE) {
+    get_list = get_autoload = get_hash = get_other = 0;
+  } else {
+    get_list = (flags & GET_LIST);
+    get_hash = (flags & GET_HASH);
+    get_autoload = (flags & GET_AUTOLOAD);
+    get_other = (get_list | get_hash | get_autoload);
+    if (get_other) {
+      flags |= GET_OPAQUE;
     }
   }
+  for (i = 0; i <= T_OPAQUE; ++i) {
+    match[i] = 0;
+  }
+  match[T_CHAR]      = (flags & GET_ARRAY);
+  match[T_SHORT]     = (flags & GET_ARRAY);
+  match[T_INT]       = (flags & GET_ARRAY);
+  match[T_LONG]      = (flags & GET_ARRAY);
+  match[T_FLOAT]     = (flags & GET_ARRAY);
+  match[T_DOUBLE]    = (flags & GET_ARRAY);
+  match[T_COMPLEX]   = (flags & GET_ARRAY);
+  match[T_STRING]    = (flags & GET_ARRAY);
+  match[T_POINTER]   = (flags & GET_ARRAY);
+  match[T_STRUCT]    = (flags & GET_STRUCT);
+  match[T_RANGE]     = (flags & GET_RANGE);
+#ifdef GET_LVALUE
+  match[T_LVALUE]    = (flags & GET_LVALUE);
+#endif
+  match[T_VOID]      = (flags & GET_VOID);
+  match[T_FUNCTION]  = (flags & GET_FUNCTION);
+  match[T_BUILTIN]   = (flags & GET_BUILTIN);
+  match[T_STRUCTDEF] = (flags & GET_STRUCTDEF);
+  match[T_STREAM]    = (flags & GET_STREAM);
+  match[T_OPAQUE]    = (flags & GET_OPAQUE);
 
-  PushDataBlock(RefNC(&nilDB));
+  /* Counter number of matching symbols. */
+  ret = NULL; /* avoids compiler warning */
+  number = 0;
+  for (pass = 0; pass <= 1; ++pass) {
+    if (pass) {
+      if (number <= 0) {
+	/* No matching symbols found. */
+	PushDataBlock(RefNC(&nilDB));
+	return;
+      }
+      ret = YETI_PUSH_NEW_Q(ynew_dim(number, NULL));
+    }
+    for (i=0 ; i<nitems ; ++i) {
+      if (globTab[i].ops != &dataBlockSym) {
+	continue;
+      }
+      ops = globTab[i].value.db->ops;
+      type = ops->typeID;
+      if ((unsigned int)type > T_OPAQUE || ! match[type]) {
+	continue;
+      }
+      if (get_other && type == T_OPAQUE) {
+	if (ops == &listOps) {
+	  if (! get_list) {
+	    continue;
+	  }
+	} else if (ops == &hashOps) {
+	  if (! get_hash) {
+	    continue;
+	  }
+	} else if (ops == &auto_ops) {
+	  if (! get_autoload) {
+	    continue;
+	  }
+	}
+      }
+      if (pass) {
+	*ret++ = p_strcpy(globalTable.names[i]);
+      } else {
+	++number;
+      }
+    }
+  }
 }
 #undef GET_ARRAY
 #undef GET_STRUCT
@@ -1116,6 +1172,9 @@ void Y_symbol_names(int argc)
 #undef GET_STRUCTDEF
 #undef GET_STREAM
 #undef GET_OPAQUE
+#undef GET_LIST
+#undef GET_HASH
+#undef GET_AUTOLOAD
 
 extern BuiltIn Y_insure_temporary;
 
@@ -1207,7 +1266,7 @@ void Y_is_matrix(int nargs)
 		dims->next && ! dims->next->next));
 }
 
-#if PROVIDE_IS_LIST
+#if YETI_MUST_PROVIDE_IS_LIST
 void Y_is_list(int nargs)
 {
   Symbol *s;
@@ -1225,7 +1284,7 @@ void Y_is_list(int nargs)
     PushIntValue(0);
   }
 }
-#endif /* PROVIDE_IS_LIST */
+#endif /* YETI_MUST_PROVIDE_IS_LIST */
 
 void Y_is_integer(int nargs)
 {
